/*
 * $QNXLicenseC: 
 * Copyright 2012, 2013 QNX Software Systems.  
 *  
 * Licensed under the Apache License, Version 2.0 (the "License"). You  
 * may not reproduce, modify or distribute this software except in  
 * compliance with the License. You may obtain a copy of the License  
 * at: http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" basis,  
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied. 
 * 
 * This file may contain contributions from others, either as  
 * contributors under the License or as licensors under other terms.   
 * Please review this entire file for other proprietary rights or license  
 * notices, as well as the QNX Development Suite License Guide at  
 * http://licensing.qnx.com/license-guide/ for other information. 
 * $
 */

#include "startup.h"
#include "mx6x_epit.h"
#include "board.h"

/* Global variables used when toggling GPIO reset lines */
uint32_t 	timer_value;
uint32_t	gpio_base;
uint32_t	gpio_pin;
uint32_t	gpio_level;

/*
 * Name:	mx6x_set_gpio_output
 * Description:	This function configures a GPIO pin as an input.
 * Parameters:
 * 		base -Base address (physical address) of GPIO bank
 *		pin - GPIO pin (0 - 32) e.g. 1<<9 for pin #9
 *		level - GPIO output value
 * Return Value: void
 */
void mx6x_set_gpio_output(uint32_t base, uint32_t pin, uint32_t level)
{
	uint32_t reg = in32(base + MX6X_GPIO_DR);

	/* Set GPIO output level */
	if (level)
		reg |= pin;
	else
		reg &= ~pin;
	out32(base + MX6X_GPIO_DR, reg);

	/* Set the GPIO pin to output mode */
	reg = in32(base + MX6X_GPIO_GDIR);
	reg |= pin;
	out32(base + MX6X_GPIO_GDIR, reg);
}

/*
 * Name:	mx6x_set_gpio_input
 * Description:	This function configures a GPIO pin as an input.
 * Parameters:
 * 		base -Base address (physical address) of GPIO bank
 *		pin - GPIO pin (0 - 32) e.g. 1<<9 for pin #9
 * Return Value: void
 */
void mx6x_set_gpio_input(uint32_t base, uint32_t pin)
{
	/* Set the GPIO pin to input mode */
	uint32_t reg = in32(base + MX6X_GPIO_GDIR);
	reg &= ~pin;
	out32(base + MX6X_GPIO_GDIR, reg);
}

/*
 * Name:	 mx6x_reset_gpio_pin
 * Description:	 This function configures a GPIO as an output, sets the output value, and sets
 *               up a timer. mx6x_reset_gpio_pin and mx6x_reset_gpio_pin_fin can be used to put a chip
 *				 in reset for a duration of time without blocking other startup code from running.
 * Parameters:
 * 		base -Base address (physical address) of GPIO bank
 *		pin - GPIO pin (0 - 32) e.g. 1<<9 for pin #9
 *		level - GPIO output value
 * Return Value: void
 */
void mx6x_reset_gpio_pin(uint32_t base, uint32_t pin, uint32_t level)
{
	/* Global variables needed by mx6x_reset_gpio_pin_fin */
	gpio_base = base;
	gpio_pin = pin;
	gpio_level = level;

	/* Configure GPIO as output, and put chip into reset */
	mx6x_set_gpio_output(base, pin, gpio_level);

	/* Start the counter */
	timer_value = mx6_epit_get_timer_val();
}

/*
 * Name:	 mx6x_reset_gpio_pin_fin
 * Description:	 This function takes a chip out of reset after a specified duration of time
 *               has elapsed.
 * Parameters:
 * 		 dur - the amount of time that the chip must be in reset, value specified
 *		      in microseconds.
 * Return Value: void
 */
void mx6x_reset_gpio_pin_fin(uint32_t dur)
{
	dur = dur * IPG_CLKS_IN_ONE_US;
	while ((timer_value - mx6_epit_get_timer_val()) <= dur)
		__asm__ __volatile__("nop");

	/* Bring chip out of reset */
	uint32_t reg = in32(gpio_base + MX6X_GPIO_DR);

	if (gpio_level)
		reg &= ~gpio_pin;
	else
		reg |= gpio_pin;

	out32(gpio_base + MX6X_GPIO_DR, reg);
}

#if defined(__QNXNTO__) && defined(__USESRCVERSION)
#include <sys/srcversion.h>
__SRCVERSION("$URL: http://svn/product/branches/6.6.0/trunk/hardware/startup/boards/imx6x/mx6x_reset_gpio_pin.c $ $Rev: 712370 $")
#endif
