#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <sys/term.h>
#include <sys/ioctl.h>

#include "fpga.h"
#include "crossbar-ts4900.h"
#include "crossbar-ts7970.h"
#include "crossbar-ts7990.h"

static int twifd;

static void auto485_en(int uart, int baud, char *mode);

static void usage(char **argv) {
	fprintf(stderr,
		"Usage: %s [OPTIONS] ...\n"
		"Technologic Systems I2C FPGA Utility\n"
		"\n"
		"  -b board       Select one of 0x4900, 0x7970, 0x7990\n"   
		"  -m <address>   Sets up the address for a peek/poke\n"
		"  -v <value>     Writes the value to the specified address\n"
		"  -t             Reads from the specified address\n"
		"  -i <8n1>       Used with -a, sets mode like '8n1', '7e2', etc\n"
		"  -x <speed>     Used with -a, sets baud rate for auto485\n"
		"  -a <uart>      Enables autotxen for supported CPU UARTs\n"
		"                    Uses baud/mode if set or reads the current\n"
		"                    configuration of that uart\n"
		"  -c             Prints out the crossbar configuration\n"
		"  -g             Print crossbar for use in eval\n"
		"  -s             Read environment for crossbar changes\n"
      "  -q             Print all possible FPGA inputs and outputs.\n"
		"  -h             This message\n"
		"\n",
		argv[0]
	);
}

int main(int argc, char **argv) 
{
	int c;
	unsigned short addr = 0x0;
	int opt_addr = 0;
	int opt_poke = 0, opt_peek = 0, opt_auto485 = 0;
	int opt_get = 0, opt_set = 0, opt_dump = 0, opt_showall = 0;
	int baud = 0;
	int model = 0;
	unsigned char pokeval = 0;
	char *uartmode = 0;
	struct cbarpin *cbar_inputs, *cbar_outputs;
	int cbar_size, cbar_mask;
	
	twifd = fpga_init();
		
	if(twifd == -1) {
		perror("Can't open FPGA I2C bus");
		return 1;
	}

	while((c = getopt(argc, argv, "+m:v:i:x:ta:b:cgsqh")) != -1) {		
		switch(c) {
      case 'b':
         model = strtoull(optarg, NULL, 0);
         if(model == 0x4900) {
		      cbar_inputs = ts4900_inputs; 
		      cbar_outputs = ts4900_outputs;
		      cbar_size = 5;
		      cbar_mask = 7;
	      } else if(model == 0x7970) {
		      cbar_inputs = ts7970_inputs; 
		      cbar_outputs = ts7970_outputs;
		      cbar_size = 6;
		      cbar_mask = 3;
	      } else if(model == 0x7990) {
		      cbar_inputs = ts7990_inputs; 
		      cbar_outputs = ts7990_outputs;
		      cbar_size = 6;
		      cbar_mask = 3;
	      } else {
		      fprintf(stderr, "Unsupported model %d\n", model);
		      return 1;
	      }
         break;
		case 'm':
			opt_addr = 1;
			addr = strtoull(optarg, NULL, 0);
			break;
		case 'v':
			opt_poke = 1;
			pokeval = strtoull(optarg, NULL, 0);
			break;
		case 'i':
			uartmode = strdup(optarg);
			break;
		case 'x':
			baud = atoi(optarg);
			break;
		case 't':
			opt_peek = 1;
			break;
		case 'a':
			opt_auto485 = atoi(optarg);
			break;
		case 'g':
		   opt_get = 1;
		   break;		   
		case 's':
		   opt_set = 1;
		   break;
		case 'c':
		   opt_dump = 1;
		   break;
		case 'q':
		   opt_showall = 1;
		   break;			
		default:
			usage(argv);
		}
	}
	
	if (model == 0) {
	   fprintf(stderr, "No model supplied!  Use the -b option\n");
	   return 1;
	}
	
	if (opt_get) {				   
	   int i;
      for (i = 0; cbar_inputs[i].name != 0; i++)
		{
			int j;
			unsigned char mode = fpeek8(twifd, cbar_inputs[i].addr) >> (8 - cbar_size);
			for (j = 0; cbar_outputs[j].name != 0; j++)
			{
				if(cbar_outputs[j].addr == mode){
					printf("%s=%s\n", cbar_inputs[i].name, cbar_outputs[j].name);
					break;
				}
			}
		}
	}

	if (opt_set) {
	   int i;
   	for (i = 0; cbar_inputs[i].name != 0; i++)
		{
			char *value = getenv(cbar_inputs[i].name);
			int j;
			if(value != NULL) {
				for (j = 0; cbar_outputs[j].name != 0; j++) {
					if(strcmp(cbar_outputs[j].name, value) == 0) {
						int mode = cbar_outputs[j].addr;
						unsigned char val = fpeek8(twifd, cbar_inputs[i].addr);
						fpoke8(twifd, cbar_inputs[i].addr, 
							   (mode << (8 - cbar_size)) | (val & cbar_mask));
							break;
					}
				}
				if(cbar_outputs[i].name == 0) {
					fprintf(stderr, "Invalid value \"%s\" for input %s\n",
						value, cbar_inputs[i].name);
				}
			} 
		}
	}

	if (opt_dump) {
	   int i;
		printf("%13s (DIR) (VAL) FPGA Output\n", "FPGA Pad");
		for (i = 0; cbar_inputs[i].name != 0; i++)
		{
			unsigned char value = fpeek8(twifd, cbar_inputs[i].addr);
			unsigned char mode = value >> (8 - cbar_size);
			char *dir = value & 0x1 ? "out" : "in";
			int val;
			int j;

			// 4900 uses 5 bits for cbar, 7970/7990 use 6 and share
			// the data bit for input/output
			if(value & 0x1 || cbar_size == 6) {
				val = value & 0x2 ? 1 : 0;
			} else {
				val = value & 0x4 ? 1 : 0;
			}
			
			for (j = 0; cbar_outputs[j].name != 0; j++)
			{
				if(cbar_outputs[j].addr == mode){
					printf("%13s (%3s) (%3d) %s\n", 
						cbar_inputs[i].name,
						dir,
						val,
						cbar_outputs[j].name);
					break;
				}
			}
		}
	}
	
	if (opt_showall) {
  	   int i;
	   printf("FPGA Inputs:\n");
		for (i = 0; cbar_inputs[i].name != 0; i++) {
			printf("%s\n", cbar_inputs[i].name);
		}
		printf("\nFPGA Outputs:\n");
		for (i = 0; cbar_outputs[i].name != 0; i++) {
			printf("%s\n", cbar_outputs[i].name);
		}
	}
	
	if(opt_poke) {
		if(opt_addr) {
			fpoke8(twifd, addr, pokeval);
		} else {
			fprintf(stderr, "No address specified\n");
		}
	}

	if(opt_peek) {
		if(opt_addr) {
			printf("addr%d=0x%X\n", addr, fpeek8(twifd, addr));
		} else {
			fprintf(stderr, "No address specified\n");
		}
	}

	if(opt_auto485) {
		if(opt_auto485 != 2 && opt_auto485 != 4) {
			fprintf(stderr, "Specify value 2 or 4 for /dev/ser2 or /dev/ser4 (specified %d)\n", opt_auto485);
			return 1;
		} 
		auto485_en(opt_auto485, baud, uartmode);
	}

	close(twifd);

	return 0;
}

// Calculate the number of 24mhz clocks for a given
// baud rate / bits per symbol
// For example:
////// CNT 1
//// 115200 is 8681ns byte time
//// 8681*9.5=82469.5ns
//// 24mhz is 41.67ns
//// 82469.5/41.67 = 1979 (CNT1)
////// CNT 2
//// 4340.5ns is .5 bit time
//// 4340.5/41.67 = 104 (CNT2)
static void autotx_bitstoclks(int bits, int baud, unsigned int *cnt1, unsigned int *cnt2)
{
#if (0)
	float clk;
	const float fpga_clk = 41.67; // ns

	// get baud byte time in ns
	clk = (bits*100000000)/baud;
	*cnt1 = lround((clk*9.5) / fpga_clk);
	*cnt2 = lround((clk*0.5) / fpga_clk);
#else
   // integer version of above */
   unsigned long i_clk;		
		
	i_clk = (100000000 / baud) * bits * 1000;			
	*cnt1 = (i_clk * 190) / 833333;
	*cnt2 = i_clk / 83333;
#endif

}  

void auto485_en(int uart, int baud, char *mode)
{
	struct termios tio;
	int uartfd;
	int symsz;
	unsigned int cnt1, cnt2;

	if(uart == 2)
		uartfd = open("/dev/ser2", O_RDONLY);
	else if (uart == 4)
		uartfd = open("/dev/ser4", O_RDONLY);

	if (tcgetattr(uartfd, &tio)) {
	   fprintf(stderr, "Error:  Can't open %s\n", uart == 2?"/dev/ser2":"/dev/ser4");
	   return;
	}  
		
	if(baud != 0 && mode != NULL) {	
		// If they did specify them, parse the symbol size
		// from the mode and set the new termios
		cfsetispeed(&tio, baud);
		cfsetospeed(&tio, baud);
		
		tio.c_cflag &= ~CSIZE;
		if(mode[0] == '8') tio.c_cflag |= CS8;
		else if(mode[0] == '7') tio.c_cflag |= CS7;
		else if(mode[0] == '6') tio.c_cflag |= CS6;
		else if(mode[0] == '5') tio.c_cflag |= CS5;
		if(mode[1] == 'n') tio.c_cflag &= ~PARENB;
		else if(mode[1] == 'e'){
			tio.c_cflag |= PARENB;
			tio.c_cflag &= ~PARODD;
		} else if (mode[1] == 'o') tio.c_cflag |= PARENB | PARODD;
		if(mode[2] == '1') tio.c_cflag &= ~CSTOPB;
		else if (mode[2] == '2') tio.c_cflag |= CSTOPB;
		if (tcsetattr(uartfd, TCSADRAIN, &tio))
		   fprintf(stderr, "Error %d in tcsetattr\n", errno);   
	}

	// If they didnt specify a mode & baud, 
	// just read the current ioctl state and 
	// go with that
		
	baud = cfgetispeed(&tio);
		
	if((tio.c_cflag & CS8) == CS8) symsz = 10;
	else if((tio.c_cflag & CS7) == CS7) symsz = 9;
	else if((tio.c_cflag & CS6) == CS6) symsz = 8;
	else if((tio.c_cflag & CS5) == CS5) symsz = 7;
	if(tio.c_cflag & CSTOPB) symsz++;
	if(tio.c_cflag & PARENB) symsz++;

	close(uartfd);

	if(mode == NULL) {
		printf("Setting Auto TXEN for %d baud and %d bits per symbol from current settings\n",
		baud, symsz);
	} else {
		printf("Setting Auto TXEN for %d baud and %d bits per symbol (%s)\n",
		baud, symsz, mode);
	}

	if(uart == 2) {
		autotx_bitstoclks(symsz, baud, &cnt1, &cnt2);
		fpoke8(twifd, 32, (unsigned char)((cnt1 & 0xff0000) >> 16));
		fpoke8(twifd, 33, (unsigned char)((cnt1 & 0xff00) >> 8));
		fpoke8(twifd, 34, (unsigned char)(cnt1 & 0xff));
		fpoke8(twifd, 35, (unsigned char)((cnt2 & 0xff0000) >> 16));
		fpoke8(twifd, 36, (unsigned char)((cnt2 & 0xff00) >> 8));
		fpoke8(twifd, 37, (unsigned char)(cnt2 & 0xff));
	} else if (uart == 4) {
		autotx_bitstoclks(symsz, baud, &cnt1, &cnt2);
		fpoke8(twifd, 38, (unsigned char)((cnt1 & 0xff0000) >> 16));
		fpoke8(twifd, 39, (unsigned char)((cnt1 & 0xff00) >> 8));
		fpoke8(twifd, 40, (unsigned char)(cnt1 & 0xff));
		fpoke8(twifd, 41, (unsigned char)((cnt2 & 0xff0000) >> 16));
		fpoke8(twifd, 42, (unsigned char)((cnt2 & 0xff00) >> 8));
		fpoke8(twifd, 43, (unsigned char)(cnt2 & 0xff));
	}	
}
